from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Tuple,
    TYPE_CHECKING,
    Union,
)

from pyvenafi.logger import features_logger
from pyvenafi.tpp.api.api_base import InvalidResponseError
from pyvenafi.tpp.attributes.x509_certificate import X509CertificateAttributes
from pyvenafi.tpp.features.bases.feature_base import (
    feature,
    FeatureBase,
)
from pyvenafi.tpp.features.definitions.classes import Classes
from pyvenafi.tpp.features.definitions.exceptions import (
    FeatureException,
    UnexpectedValue,
)

if TYPE_CHECKING:
    from pyvenafi.tpp.api.websdk.models import (
        config,
        identity as ident,
    )

StringParam = Union[str, list[str], list[Tuple[str, ...]]]

@dataclass
class DownloadedCertificate:
    certificate_data: str
    filename: str
    format: str

@feature('Certificate')
class Certificate(FeatureBase):
    def __init__(self, api):
        super().__init__(api)

    def _get(self, certificate: 'Union[config.Object, str]'):
        # High volume concurrency can cause a 500 internal error in IIS due to a deadlock.
        # In this case the error requests the client to "rerun the transaction".
        certificate_guid = self._get_guid(certificate)

        retries = 3
        while retries > 0:
            try:
                result = self._api.websdk.Certificates.Guid(certificate_guid).get()
                if 'Rerun the transaction' not in result.api_response.reason:
                    return result
            except InvalidResponseError as err:
                if 'Rerun the transaction' in err.response.reason:
                    # A deadlock issue may have caused this. Try getting the certificate again.
                    features_logger.debug('Rerunning Certificates/Get transaction due to deadlock...')
            retries -= 1
        return self._api.websdk.Certificates.Guid(certificate_guid).get()

    def associate_application(
        self, certificate: 'Union[config.Object, str]', applications: 'list[Union[config.Object, str]]',
        push_to_new: bool = False
    ):
        """
        Associates an application object to a certificate object.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` for each application object.
            push_to_new: If ``True``, the certificate will be pushed to the application once associated.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Associate.post(
            application_dn=[self._get_dn(a) for a in applications],
            certificate_dn=certificate_dn,
            push_to_new=push_to_new
        )
        if not result.success:
            raise FeatureException(f'Unable to associate the given applications to the certificate "{certificate_dn}".')

    def create(
        self,
        name: str,
        parent_folder: 'Union[config.Object, str]',
        description: 'str' = None,
        contacts: 'list[Union[ident.Identity, str]]' = None,
        approvers: 'list[Union[ident.Identity, str]]' = None,
        config_class: 'str' = Classes.x509_certificate,
        management_type: 'str' = None,
        service_generated_csr: 'bool' = None,
        generate_key_on_application: 'bool' = None,
        hash_algorithm: 'str' = None,
        common_name: 'str' = None,
        organization: 'str' = None,
        organization_unit: 'list[str]' = None,
        city: 'str' = None,
        state: 'str' = None,
        country: 'str' = None,
        san_dns: 'list[str]' = None,
        san_email: 'list[str]' = None,
        san_upn: 'list[str]' = None,
        san_ip: 'list[str]' = None,
        san_uri: 'list[str]' = None,
        key_algorithm: 'str' = None,
        key_strength: 'int' = None,
        elliptic_curve: 'str' = None,
        ca_template: 'Union[config.Object, str]' = None,
        disable_automatic_renewal: 'bool' = None,
        renewal_window: 'int' = None,
        attributes: dict = None,
        get_if_already_exists: bool = True
    ):
        """
        .. note::
            The certificate is not automatically requested. Use :meth:`renew` to obtain a certificate.

        Args:
            name: Name of the Certificate object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            description: Description of the certificate object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` of the contacts for this certificate.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` of the approvers for this certificate.
            management_type: Certificate management type.
            service_generated_csr: If ``True``, TPP generates the CSR.
            generate_key_on_application: If ``True``, the key/CSR are generated on the target application.
            hash_algorithm: Hash algorithm.
            common_name: Common name (CN) of the certificate.
            organization: Organization of the certificate.
            organization_unit: Organization units (OU) of the certificate.
            city: City.
            state: State.
            country: Country code.
            san_dns: List of Subject Alternative Names for DNS.
            san_email: List of Subject Alternative Names for e-mail addresses.
            san_upn: List of Subject Alternative Names for UPN.
            san_ip: List of Subject Alternative Names for IP addresses.
            san_uri: List of Subject Alternative Names for URI.
            key_algorithm: Signing key algorithm.
            key_strength: Key strength in bits.
            elliptic_curve: Elliptic curve.
            ca_template: :ref:`config_object` or :ref:`dn` of the Certificate Authority template.
            disable_automatic_renewal: If ``True``, disables automatic renewal.
            renewal_window: Number of days that make the renewal window.
            attributes: Additional attributes that define this certificate.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object`
        """
        cert_attrs = {
            X509CertificateAttributes.description                      : description,
            X509CertificateAttributes.contact                          : [self._get_prefixed_universal(c) for c in
                                                                          contacts] if contacts else None,
            X509CertificateAttributes.approver                         : [self._get_prefixed_universal(a) for a in
                                                                          approvers] if approvers else None,
            X509CertificateAttributes.management_type                  : management_type,
            X509CertificateAttributes.manual_csr                       : {
                True : "0",
                False: "1"
            }.get(service_generated_csr),
            X509CertificateAttributes.generate_keypair_on_application  : {
                True : "1",
                False: "0"
            }.get(generate_key_on_application),
            X509CertificateAttributes.pkcs10_hash_algorithm            : hash_algorithm,
            X509CertificateAttributes.x509_subject                     : common_name,
            X509CertificateAttributes.organization                     : organization,
            X509CertificateAttributes.organizational_unit              : organization_unit,
            X509CertificateAttributes.city                             : city,
            X509CertificateAttributes.state                            : state,
            X509CertificateAttributes.country                          : country,
            X509CertificateAttributes.driver_name                      : 'appx509certificate',
            X509CertificateAttributes.x509_subjectaltname_dns          : san_dns,
            X509CertificateAttributes.x509_subjectaltname_ipaddress    : san_ip,
            X509CertificateAttributes.x509_subjectaltname_uri          : san_uri,
            X509CertificateAttributes.x509_subjectaltname_othername_upn: san_upn,
            X509CertificateAttributes.x509_subjectaltname_rfc822       : san_email,
            X509CertificateAttributes.key_algorithm                    : key_algorithm,
            X509CertificateAttributes.key_bit_strength                 : key_strength,
            X509CertificateAttributes.elliptic_curve                   : elliptic_curve,
            X509CertificateAttributes.certificate_authority            : self._get_dn(
                ca_template
            ) if ca_template else None,
            X509CertificateAttributes.disable_automatic_renewal        : {
                True : "1",
                False: "0"
            }.get(disable_automatic_renewal),
            X509CertificateAttributes.renewal_window                   : renewal_window
        }
        if attributes:
            cert_attrs.update(attributes)
        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=config_class,
            attributes=cert_attrs,
            get_if_already_exists=get_if_already_exists
        )

    def delete(self, certificate: 'Union[config.Object, str]'):
        """
        Deletes the certificate object from TPP.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_guid = self._get_guid(certificate)
        try:
            self._api.websdk.Certificates.Guid(certificate_guid).delete()
        except InvalidResponseError:
            raise FeatureException(f'Could not delete certificate {str(certificate)}.')

    def details(self, certificate: 'Union[config.Object, str]'):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.

        Returns:
            :class:`~.models.certificate.CertificateDetails`
        """
        return self._get(certificate=certificate).certificate_details

    def dissociate_application(
        self, certificate: 'Union[config.Object, str]', applications: 'list[Union[config.Object, str]]',
        delete_orphans: bool = False
    ):
        """
        Dissociate an application object from a certificate.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` to each application object.
            delete_orphans: Delete the application object. The corresponding device object will be delete if it has no
                            child applications. Use this option to completely remove the application object and corresponding
                            device objects.
        """
        certificate_dn = self._get_dn(certificate)
        self._api.websdk.Certificates.Dissociate.post(
            certificate_dn=certificate_dn,
            application_dn=[self._get_dn(a) for a in applications],
            delete_orphans=delete_orphans
        )

    # noinspection ALL
    def download(
        self, format: str, certificate: 'Union[config.Object, str]' = None, friendly_name: str = None,
        include_chain: bool = False, include_private_key: bool = False, keystore_password: str = None,
        password: str = None, root_first_order: bool = False, vault_id: int = None, timeout: int = 60,
        poll_interval: float = 0.5
    ):
        """
        Downloads a certificate and returns the encoded content, filename, and format as a single object. If ``vault_id``
        is provided, then that specific version of a certificate is downloaded, which is particularly useful when
        trying to download historical certificates.

        Attempts to download the certificate occur every ``poll_interval`` seconds for up to a ``timeout`` threshold. This
        is because the certificate may be currently processing and not ready for download.

        Args:
            format: One of the following:
                * Base64
                * Base64 (PKCS #8)
                * DER
                * JKS
                * PKCS #7
                * PKCS #12
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object. Not required if using vault id.
            friendly_name: Label or alias for the given format.
            include_chain: Include parent or root chain.
            include_private_key: Include the private key.
            keystore_password: JKS Keystore password.
            password: Password.
            root_first_order: If ``True``, show root certificate first, followed by intermediate, and finally the
                end entity certificate.
            vault_id: If provided, downloads the certificate with the given Vault ID. Use this when trying
                to download historical certificates. Not required if using certificate config object.
            timeout: Timeout threshold in seconds until a TimeoutError is raised.
            poll_interval: Time in seconds to attempt a retry of downloading the certificate.

        Returns:
            A *DownloadedCertificate* with these properties

            * **certificate_data** *str* - Encoded certificate data.
            * **format** *str* - File Format.
            * **filename** *str* - File name.
        """
        exception = None
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired():
                try:
                    if vault_id:
                        result = self._api.websdk.Certificates.Retrieve.VaultId(vault_id).post(
                            format=format,
                            friendly_name=friendly_name,
                            include_chain=include_chain,
                            include_private_key=include_private_key,
                            keystore_password=keystore_password,
                            password=password,
                            root_first_order=root_first_order
                        )

                    else:
                        certificate_dn = self._get_dn(certificate)
                        result = self._api.websdk.Certificates.Retrieve.post(
                            certificate_dn=certificate_dn,
                            format=format,
                            friendly_name=friendly_name,
                            include_chain=include_chain,
                            include_private_key=include_private_key,
                            keystore_password=keystore_password,
                            password=password,
                            root_first_order=root_first_order
                        )

                    return DownloadedCertificate(
                        certificate_data=result.certificate_data,
                        filename=result.filename,
                        format=result.format
                    )
                except Exception as e:
                    exception = e
                to.poll(poll_interval)
        if exception:
            raise TimeoutError('Failed to download the certificate.') from exception
        raise TimeoutError('Failed to download the certificate.')

    def get(self, certificate_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            certificate_dn: :ref:`dn` of the certificate object.
            raise_error_if_not_exists: Raise an exception if the object DN does not exist.

        Returns:
            :ref:`config_object`
        """
        return self._get_config_object(object_dn=certificate_dn, raise_error_if_not_exists=raise_error_if_not_exists)

    def get_previous_versions(
        self, certificate: 'Union[config.Object, str]', exclude_expired: bool = False,
        exclude_revoked: bool = False
    ):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            exclude_expired: If ``True``, do not include expired certificates.
            exclude_revoked: If ``True``, do not include revoked certificates.

        Returns:
            list[:class:`~.models.certificate.PreviousVersions`]
        """
        certificate_guid = self._get_guid(certificate)
        result = self._api.websdk.Certificates.Guid(certificate_guid).PreviousVersions.get(
            exclude_expired=exclude_expired,
            exclude_revoked=exclude_revoked
        )
        return result.previous_versions

    def get_validation_results(self, certificate: 'Union[config.Object, str]'):
        """
        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.

        Returns:
            Tuple[:class:`~.models.certificate.File`, :class:`~.models.certificate.SslTls`]
        """
        certificate_guid = self._get_guid(certificate)
        result = self._api.websdk.Certificates.Guid(certificate_guid).ValidationResults.get()
        return result.file, result.ssl_tls

    def list(
        self,
        country: StringParam = None,
        certificate_type: StringParam = None,
        common_name: StringParam = None,
        issuer: StringParam = None,
        is_self_signed: bool = None,
        is_wild_card: bool = None,
        key_algorithm: StringParam = None,
        key_size: StringParam = None,
        key_size_greater: StringParam = None,
        key_size_less: StringParam = None,
        city: StringParam = None,
        organization: StringParam = None,
        organization_unit: StringParam = None,
        state: StringParam = None,
        san_dns: StringParam = None,
        san_email: StringParam = None,
        san_ip: StringParam = None,
        san_upn: StringParam = None,
        san_uri: StringParam = None,
        serial: StringParam = None,
        signature_algorithm: StringParam = None,
        thumbprint: StringParam = None,
        valid_from: Union[datetime, str] = None,
        valid_from_greater: Union[datetime, str] = None,
        valid_from_less: Union[datetime, str] = None,
        valid_to: Union[datetime, str] = None,
        valid_to_greater: Union[datetime, str] = None,
        chain_validation_error: StringParam = None,
        created_on: Union[datetime, str] = None,
        created_on_greater: Union[datetime, str] = None,
        created_on_less: Union[datetime, str] = None,
        disabled: bool = None,
        in_error: bool = None,
        management_type: StringParam = None,
        name: StringParam = None,
        network_validation_disabled: bool = None,
        parent_folder: 'Union[config.Object, str]' = None,
        recursive: bool = True,
        pending_workflow: bool = None,
        ssl_tls_protocol: StringParam = None,
        stage: StringParam = None,
        stage_greater: str = None,
        stage_less: str = None,
        tls_validation_failure: StringParam = None,
        validation_disabled: bool = None,
        validation_state: StringParam = None,
        valid_to_less: Union[datetime, str] = None,
        additional_filters: dict[str, str] = None,
        limit: int = 1000,
        offset: int = None,
        return_limit: bool = False,
        include_issuer: bool = None,
        include_key_algorithm: bool = None,
        include_key_size: bool = None,
        include_subject: bool = None,
        concurrency: int = 16,
        cap: int = None
    ):
        """
        Lists all certificates with the given parameters. Some parameters allow a union of values:

        * If ``datetime``, then the ISO format is used.
        * If a list of strings, then each value is joined by an ampersand (&) and appended to the query string.
        * If a list of a tuple of strings, then each value in the tuple is joined by a comma (,) and each tuple in
          the list is joined by an ampersand (&), and the whole joined to the query string.

        Refer to the |Doc Home Page| for detailed parameter descriptions.

        Args:
            country: Country attribute of the Subject DN.
            certificate_type: Use CodeSigning, Device, Server, and/or User.
            common_name: CN of the Subject DN.
            issuer: Issuer.
            is_self_signed: Include only self-signed (``True``) or CA-signed (``False``) certificates.
            is_wild_card: Include only certificates that have wild card notation (``True``) or those that do not (``False``).
            key_algorithm: Public key algorithm.
            key_size: Public key size.
            key_size_greater: Minimum key size, non-inclusive.
            key_size_less: Maximum key size, inclusive.
            city: City attribute of the Subject DN.
            organization: Organization attribute of the Subject DN.
            organization_unit: Organization Unit attribute of the Subject DN.
            state: State attribute of the Subject DN.
            san_dns: SAN DNS of the certificate.
            san_email: SAN E-mail of the certificate.
            san_ip: SAN IP of the certificate.
            san_upn: SAN UPN of the certificate.
            san_uri: SAN URI of the certificate.
            serial: Serial number of the certificate.
            signature_algorithm: Signature algorithm used to sign the certificate.
            thumbprint: Certificate thumbprint.
            valid_from: Issuance date of the certificate.
            valid_from_greater: Minimum expiration date of the certificate.
            valid_from_less: Maximum expiration date of the certificate.
            valid_to: Expiration date of the certificate.
            valid_to_greater: Minimum expiration date of the certificate.
            valid_to_less: Maximum expiration date of the certificate.
            chain_validation_error: Chain Validation Error code.
            created_on: Creation date of the certificate.
            created_on_greater: Minimum creation date of the certificate.
            created_on_less: Maximum creation date of the certificate.
            disabled: Include only enabled (``False``) or disabled (``True``) certificates.
            in_error: Include only certificates in an error state (``True``) or a healthy state (``False``).
            management_type: Management type (Unassigned, Monitoring, Enrollment, or Provisioning) of the certificate.
            name: Name of the certificate object.
            network_validation_disabled: Include only certificates with Network Validation enabled (``False``) or disabled (``True``).
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            recursive: Requires ``parent`` to be given. If ``True``, search for certificates recursively.
            pending_workflow: Include only certificates pending workflow resolution.
            ssl_tls_protocol: SSL/TLS Protocols (Ssl2, Ssl3, Tls, Tls11, Tls12) that failed to communicate with the target host.
            stage: Processing stage of the certificate.
            stage_greater: Minimum processing stage of the certificate.
            stage_less: Maximum processing stage of the certificate.
            tls_validation_failure: TLS Validation Failure code.
            validation_disabled: Include only certificates with validation enabled (``False``) or disabled (``True``).
            validation_state: Include only certificates having a validation state of "Blank", "Success", or "Failure".
            additional_filters: A dictionary of additional filters as would be passed directly to the GET Certificates API.
            limit: The maximum number of certificates to return.
            offset: The offset from the filtered list that will be the first item of the returned list of certificates.
                Obsolete if ``return_limit`` is ``False``.
            include_issuer: Show the DN of the issuing CA.
            include_key_algorithm: Show the public key algorithm.
            include_key_size: Show the public key size.
            include_subject: Show the certificate subject DN.
            return_limit: If ``True``, then only one API call is sent to the server and the amount of certificates up to the
                ``limit`` will be returned. If ``False`` then return all certificates that apply to the given filters.
            concurrency: When return_limit is ``False``, then this is the amount of threads to spawn when querying for certificates.
                This is done by first submitting a HEAD Certificates request to get the count of certificates that apply
                to the filters and then by creating a thread pool for as many offsets are required to retrieve the total amount
                of certificates.
            cap: When the desired amount of certificates exceeds the desired limit, the cap defines how many certificates from the
                offset and beyond the limit to retrieve. For example, with ``offset=10``, ``limit=200``, and ``cap=400``, then
                certificates 10 thru 410 will be returned at a rate of 200 retrievals.

        Returns:
            list[:class:`~.models.certificate.Certificate`]
        """
        # region Filters
        filters = {
            "C"                        : country,
            "CertificateType"          : certificate_type,
            "CN"                       : common_name,
            "Issuer"                   : issuer,
            "IsSelfSigned"             : is_self_signed,
            "IsWildcard"               : is_wild_card,
            "KeyAlgorithm"             : key_algorithm,
            "KeySize"                  : key_size,
            "KeySizeGreater"           : key_size_greater,
            "KeySizeLess"              : key_size_less,
            "L"                        : city,
            "O"                        : organization,
            "OU"                       : organization_unit,
            "S"                        : state,
            "SAN-DNS"                  : san_dns,
            "SAN-Email"                : san_email,
            "SAN-IP"                   : san_ip,
            "SAN-UPN"                  : san_upn,
            "SAN-URI"                  : san_uri,
            "Serial"                   : serial,
            "SignatureAlgorithm"       : signature_algorithm,
            "Thumbprint"               : thumbprint,
            "ValidFrom"                : valid_from,
            "ValidFromGreater"         : valid_from_greater,
            "ValidFromLess"            : valid_from_less,
            "ValidTo"                  : valid_to,
            "ValidToGreater"           : valid_to_greater,
            "ValidToLess"              : valid_to_less,
            "ChainValidationFailure"   : chain_validation_error,
            "CreatedOn"                : created_on,
            "CreatedOnGreater"         : created_on_greater,
            "CreatedOnLess"            : created_on_less,
            "Disabled"                 : disabled,
            "InError"                  : in_error,
            "ManagementType"           : management_type,
            "Name"                     : name,
            "NetworkValidationDisabled": network_validation_disabled,
            "ParentDn"                 : self._get_dn(parent_folder) if parent_folder and not recursive else None,
            "ParentDnRecursive"        : self._get_dn(parent_folder) if parent_folder and recursive else None,
            "PendingWorkflow"          : pending_workflow,
            "SslTlsProtocol"           : ssl_tls_protocol,
            "Stage"                    : stage,
            "StageGreater"             : stage_greater,
            "StageLess"                : stage_less,
            "TlsValidationFailure"     : tls_validation_failure,
            "ValidationDisabled"       : validation_disabled,
            "ValidationState"          : validation_state,
        }

        def format_value(v):
            if isinstance(v, datetime):
                return v.isoformat()
            if isinstance(v, bool):
                return int(v)
            return v

        for key, value in filters.items():
            if isinstance(value, list):
                filters[key] = [
                    ",".join(format_value(i) for i in item)
                    if isinstance(item, (tuple, list))
                    else format_value(item)
                    for item in value
                ]
            else:
                filters[key] = format_value(value)
        filters.update(additional_filters or {})
        # endregion Filters

        # region Optional Fields
        optional_fields = [k for k, v in {
            "Issuer"      : include_issuer,
            "KeyAlgorithm": include_key_algorithm,
            "KeySize"     : include_key_size,
            "Subject"     : include_subject
        }.items() if v is True]
        # endregion Optional Fields

        response = self._api.websdk.Certificates.get(
            filters=filters,
            limit=limit,
            offset=offset,
            optional_fields=optional_fields
        )
        if return_limit is True or limit >= (response.total_count or 0):
            if cap and len(response.certificates) >= cap:
                return response.certificates[:cap]
            return response.certificates
        if cap and len(response.certificates) >= cap:
            return response.certificates[:cap]

        certificates = response.certificates
        total_count = min(cap, response.total_count) if cap else response.total_count
        offsets = list(range(limit, total_count, limit))

        with ThreadPoolExecutor(max_workers=concurrency) as pool:
            results = list(
                pool.map(
                    lambda o: self._api.websdk.Certificates.get(
                        filters=filters,
                        limit=limit,
                        offset=o,
                        optional_fields=optional_fields
                    ),
                    offsets
                )
            )
        for result in results:
            certificates += result.certificates

        # Keep these comments in case we need to approach this a different way.
        # from threading import Lock
        #
        # COUNTER_LOCK = Lock()
        # CERTIFICATES_LOCK = Lock()
        # current_offset = 0
        #
        # def _get():
        #     nonlocal current_offset, certificates
        #
        #     while True:
        #         with COUNTER_LOCK:
        #             o = current_offset
        #             current_offset += limit
        #         try:
        #             r = self._api.websdk.Certificates.get(
        #                 filters=filters,
        #                 limit=limit,
        #                 offset=o,
        #                 optional_fields=optional_fields
        #             )
        #             with CERTIFICATES_LOCK:
        #                 certificates += r.certificates
        #             if len(r.certificates) < limit:
        #                 return
        #         except:
        #             return
        #
        # with ThreadPoolExecutor(max_workers=concurrency) as pool:
        #     for _ in range(concurrency):
        #         pool.submit(_get)

        return certificates

    def push_to_applications(
        self,
        certificate: 'Union[config.Object, str]',
        applications: 'list[Union[config.Object, str]]' = None
    ):
        """
        Pushes the active ``certificate`` to the ``applications``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            applications: A list of :ref:`config_object` or :ref:`dn` to each application object.
        """
        if not applications:
            certificate_dn = self._get_dn(certificate)
            applications = self._api.websdk.Config.ReadDn.post(
                object_dn=certificate_dn,
                attribute_name=X509CertificateAttributes.consumers
            ).values

        self.associate_application(
            certificate=certificate,
            applications=applications,
            push_to_new=True
        )

    def renew(self, certificate: 'Union[config.Object, str]', csr: str = None, re_enable: bool = False):
        """
        Renews or requests a certificate.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            csr: If provided, uploads the PKCS10 CSR to TPP to send to the CA. If not provided, TPP generates the CSR.
            re_enable: The action to control a previously disabled certificate:
                If ``False``, do not renew a previously disabled certificate.
                If ``True``, clear the Disabled attribute, re-enable, and then renew the certificate (in this request).

        Returns:
            str: The current thumbprint of the active certificate. This should be used when checking the renewal status \
                 to ensure that TPP has registered a new certificate to its vault with a new thumbprint.
        """
        result = self._get(certificate)
        current_thumbprint = result.certificate_details.thumbprint
        self._api.websdk.Certificates.Renew.post(certificate_dn=result.dn, pkcs10=csr, reenable=re_enable)
        return current_thumbprint

    def reset(self, certificate: 'Union[config.Object, str]'):
        """
        Resets the certificate to a non-processing state. No attempt to reprocess the certificate renewal is made.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Reset.post(certificate_dn=certificate_dn, restart=False)
        if not result.processing_reset_completed:
            raise UnexpectedValue(f'Processing reset was not completed for {certificate_dn}.')

    def retry_from_current_stage(self, certificate: 'Union[config.Object, str]'):
        """
        Retries renewal from the current processing stage of the ``certificate``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        self._api.websdk.Certificates.Retry.post(certificate_dn=certificate_dn)

    def retry_from_stage_0(self, certificate: 'Union[config.Object, str]'):
        """
        Retries renewal from stage 0. This clears all current processing data and restarts
        processing.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Reset.post(certificate_dn=certificate_dn, restart=True)
        if not result.restart_completed:
            raise UnexpectedValue(f'Restart renewal from stage 0 was not triggered on {certificate_dn}.')

    def revoke(
        self, certificate: 'Union[config.Object, str]', comments: str = None, disable: bool = None,
        reason: int = None, thumbprint: str = None
    ):
        """
        Revokes the ``certificate``. If a thumbprint is provided, then the particular historical certificate
        associated to the certificate having that thumbprint will be revoked.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            comments: Any comments to include in the revoke request.
            disable: If ``True``, disables the certificate object.
            reason: Reason for revoking.
            thumbprint: If given, the thumbprint of the historical certificate to be revoked.
        """
        certificate_dn = self._get_dn(certificate)
        result = self._api.websdk.Certificates.Revoke.post(
            certificate_dn=certificate_dn,
            comments=comments,
            disable=disable,
            reason=reason,
            thumbprint=thumbprint
        )
        if result.success is not True:
            raise UnexpectedValue(
                f'Cannot revoke {certificate_dn} due to this error:\n{result.error}.'
            )

    def upload(
        self,
        certificate_data: str,
        parent_folder: 'Union[config.Object, str]',
        certificate_authority_attributes: dict = None,
        name: str = None,
        password: str = None,
        private_key_data: str = None,
        reconcile: bool = False
    ):
        """
        Uploads the certificate data to TPP to create a certificate object under the given parent folder DN. If the BEGIN/END
        header or footer is missing, the data is assumed to be Base 64 encoded in the PKCS#12 format. For Base 64 encoded
        certificates, characters, such as spaces and new line escape characters (/n), are optional. White space characters are
        removed before any attempt is made to decode the certificate.

        Args:
            certificate_data: Encoded certificate data.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            certificate_authority_attributes: Attributes pertaining to the Certificate Authority to store with the certificate
                object. This is not a DN to a Certificate Authority in TPP.
            name: If given, the name of the new certificate object. If not given, then the Common Name is used.
            password: Password to decrypt the private key.
            private_key_data: Encoded private key data.
            reconcile: If ``False``, replaces the current certificate, if it exists, and stores the current certificate as a
                historical certificate. If ``True``, then TPP activates the certificate with the newest "ValidFrom" date and
                archives the other certificate as a historical certificate.

        Returns:
            :ref:`config_object` of the uploaded certificate.
        """
        if certificate_authority_attributes:
            certificate_authority_attributes = self._name_value_list(attributes=certificate_authority_attributes)

        result = self._api.websdk.Certificates.Import.post(
            certificate_data=certificate_data,
            policy_dn=self._get_dn(parent_folder),
            ca_specific_attributes=certificate_authority_attributes,
            object_name=name,
            password=password,
            private_key_data=private_key_data,
            reconcile=reconcile
        )
        return self._api.websdk.Config.IsValid.post(object_dn=result.certificate_dn).object

    def validate(self, certificates: 'list[Union[config.Object, str]]'):
        """
        Performs SSL/TLS network validation of certificate on all applications associated to certificate that are not disabled.

        Args:
            certificates: List of :ref:`config_object` or :ref:`dn` to validate.

        Returns:
            Tuple[str, list[str]]: Tuple of :ref:`dn` and validation warnings
        """
        certificate_dns = [self._get_dn(certificate) for certificate in certificates]
        result = self._api.websdk.Certificates.Validate.post(certificate_dns=certificate_dns)
        if not result.success:
            raise UnexpectedValue(
                f'TPP failed to perform validation. Got this error: "{result.error}".'
            )
        return result.validated_certificate_dns, result.warnings

    def wait_for_enrollment_to_complete(
        self, certificate: 'Union[config.Object, str]', current_thumbprint: str,
        timeout: int = 60, poll_interval: float = 0.5
    ):
        """
        Waits for the certificate renewal to complete over a period of ``timeout`` seconds. The ``current_thumbprint``
        is returned by :meth:`renew`. Renewal is complete when the ``current_thumbprint`` does not match the new
        thumbprint and either the processing stage is "None" or greater than or equal to 800, which is the start of the
        provisioning stage. If the certificate management type is set to *Provisioning*, use the application feature
        ``wait_for_installation_to_complete()``.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            current_thumbprint: Thumbprint of the `current` certificate object.
            timeout: Timeout in seconds before raising an error.
            poll_interval: Interval to poll TPP for the renewal status.

        Returns:
            :class:`~.models.certificate.CertificateDetails`
        """
        cert = self._get(certificate=certificate)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=poll_interval):
                thumbprint = cert.certificate_details.thumbprint
                if thumbprint and thumbprint != current_thumbprint and cert.processing_details.stage in {None, 800}:
                    return cert.certificate_details
                elif cert.processing_details.in_error:
                    break
                cert = self._get(certificate=certificate)

        raise TimeoutError(
            f'Certificate renewal for "{cert.dn}" encountered an error at stage {cert.processing_details.stage} with '
            f'status "{cert.processing_details.status}".'
        )

    def wait_for_stage(
        self, certificate: 'Union[config.Object, str]', stage: int, expect_workflow: bool = True,
        timeout: int = 60, poll_interval: int = 1
    ):
        """
        Waits for the current processing of the certificate to reach the given ``stage`` over a period of
        ``timeout`` seconds. If the timeout is reached, an error is raised.

        Args:
            certificate: :ref:`config_object` or :ref:`dn` of the certificate object.
            stage: Stage at which to return
            expect_workflow: If ``True``, validates that a Ticket DN has been issued to the certificate.
            timeout: Timeout in seconds before throwing an error.
            poll_interval: Interval to poll TPP for the renewal status.

        Returns:
            The values returned by the |Websdk|, namely

            * **approver** *(list[str])* - List of approvers on the certificate object.
            * **certificate_details** (:class:`~.models.certificate.CertificateDetails`) - Certificate details.
            * **contact** *(list[str])* - List of contacts on the certificate object.
            * **created_on** *(datetime)* - Date on which the certificate object was created.
            * **custom_fields** *(list[dict])* - Custom fields on the certificate object.
            * **dn** *(str)* - :ref:`dn` of the certificate object.
            * **guid** *(str)* - :ref:`guid` of the certificate object.
            * **name** *(str)* - Name of the certificate object.
            * **parent_dn** *(str)* - Parent :ref:`dn` of the certificate object.
            * **processing_details** (:class:`~.models.certificate.ProcessingDetails`) - Certificate processing details.
            * **renewal_details** (:class:`~.models.certificate.RenewalDetails`) - Certificate renewal settings details.
            * **schema_class** *(str)* - Schema class.
            * **validation_details** (:class:`~.models.certificate.ValidationDetails`) - Certificate validation details.
        """
        cert = self._get(certificate=certificate)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired(poll=poll_interval):
                if cert.processing_details.in_error is True:
                    break
                if cert.processing_details.stage is not None:
                    if expect_workflow and cert.processing_details.stage == stage:
                        if 'pending workflow resolution' in cert.processing_details.status.lower():
                            return cert
                    elif not expect_workflow and cert.processing_details.stage >= stage:
                        return cert
                cert = self._get(certificate=certificate)

        if expect_workflow and stage == cert.processing_details.stage:
            msg = f'After {timeout} seconds certificate renewal reached stage {stage}, but no workflow was issued. ' \
                  f'The current stage is {cert.processing_details.stage} with status {cert.processing_details.status}.'
        else:
            msg = f'Certificate renewal did not reach stage {stage} after {timeout} seconds. The current ' \
                  f'stage is {cert.processing_details.stage} with status {cert.processing_details.status}.'
        raise UnexpectedValue(msg)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes

class F5LTMAdvancedTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "F5 LTM Advanced Trust Store"
    certificate_name = Attribute('Certificate Name', min_version='21.4')
    ssh_port = Attribute('SSH Port', min_version='21.4')
    use_rest_api = Attribute('Use REST API', min_version='22.1')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class SAMLAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "SAML"
    default_profile = Attribute('Default Profile')
    logout_url = Attribute('Logout Url')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class LogRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Log Root"
    log_application_container = Attribute('Log Application Container', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.log_channel import LogChannelAttributes

class LogSQLChannelAttributes(LogChannelAttributes, metaclass=IterableMeta):
    __config_class__ = "Log SQL Channel"
    create_sql_expression = Attribute('Create SQL Expression', min_version='21.4')
    dsn = Attribute('DSN', min_version='21.4')
    dsn_vault_id = Attribute('DSN Vault Id', min_version='21.4')
    database = Attribute('Database', min_version='21.4')
    expiration = Attribute('Expiration', min_version='21.4')
    expire_sql_expression = Attribute('Expire SQL Expression', min_version='21.4')
    last_run = Attribute('Last Run', min_version='21.4')
    max_log_age_days = Attribute('Max Log Age Days', min_version='21.4')
    table_name = Attribute('Table Name', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')
    view_dsn = Attribute('View DSN', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class LogNotificationAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Log Notification"
    factory_rule = Attribute('Factory Rule')
    log_channel = Attribute('Log Channel')
    rule = Attribute('Rule')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.exclusion import ExclusionAttributes

class DiscoveryExclusionAttributes(ExclusionAttributes, metaclass=IterableMeta):
    __config_class__ = "Discovery Exclusion"
    address_range = Attribute('Address Range')
    managed_by_discovery_dn = Attribute('Managed By Discovery DN')
    managed_dn = Attribute('Managed DN')
    port_range = Attribute('Port Range')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.client_work_base import ClientWorkBaseAttributes
from pyvenafi.tpp.attributes.x509_certificate_base import X509CertificateBaseAttributes

class ClientCertificateWorkAttributes(ClientWorkBaseAttributes, X509CertificateBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Client Certificate Work"
    certificate_container = Attribute('Certificate Container')
    naming_pattern = Attribute('Naming Pattern')
    transfer_allowed = Attribute('Transfer Allowed')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.client_group_base import ClientGroupBaseAttributes

class ClientSubgroupAttributes(ClientGroupBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Client Subgroup"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.flow_action_base import FlowActionBaseAttributes

class CreateKeyActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Create Key Action"
    algorithm = Attribute('Algorithm', min_version='21.4')
    key_storage_location = Attribute('Key Storage Location', min_version='21.4')
    key_usage = Attribute('Key Usage', min_version='24.1')
    vault_owner = Attribute('Vault Owner', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.schedule_base import ScheduleBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class CAImportAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "CA Import"
    amazon_private_ca = Attribute('Amazon Private CA')
    assign_contact_value_to_issued_to = Attribute('Assign Contact value to Issued To')
    automatically_import = Attribute('Automatically Import')
    aws_credential_dn = Attribute('Aws Credential DN')
    bypass_reconciliation = Attribute('Bypass Reconciliation')
    ca_type = Attribute('CA Type')
    certificate_credential = Attribute('Certificate Credential')
    certificate_origin = Attribute('Certificate Origin')
    credential = Attribute('Credential')
    default_container = Attribute('Default Container')
    discover_all = Attribute('Discover All')
    driver_name = Attribute('Driver Name')
    f5_ltm_advanced_hostname = Attribute('F5 LTM Advanced:Hostname')
    import_summary = Attribute('Import Summary')
    in_progress = Attribute('In Progress')
    include_expired = Attribute('Include Expired')
    include_revoked = Attribute('Include Revoked')
    last_processed_id = Attribute('Last Processed Id')
    last_run = Attribute('Last Run')
    last_update = Attribute('Last Update')
    log_debug = Attribute('Log Debug')
    microsoft_ca_hostname = Attribute('Microsoft CA:Hostname')
    microsoft_ca_query_by_template_name = Attribute('Microsoft CA:Query By Template Name')
    microsoft_ca_request_id = Attribute('Microsoft CA:Request ID')
    microsoft_ca_request_id_by_template_name = Attribute('Microsoft CA:Request ID By Template Name')
    microsoft_ca_service_name = Attribute('Microsoft CA:Service Name')
    microsoft_ca_template = Attribute('Microsoft CA:Template')
    page_size = Attribute('Page Size')
    placement_rule = Attribute('Placement Rule')
    placement_rule_item = Attribute('Placement Rule Item')
    placement_summary = Attribute('Placement Summary')
    powershell_script = Attribute('PowerShell Script')
    powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id')
    profile_string = Attribute('Profile String')
    region_code = Attribute('Region Code')
    result_count = Attribute('Result Count')
    s3_bucket = Attribute('S3 Bucket')
    script_execution_timeout = Attribute('Script Execution Timeout')
    script_hash_mismatch_error = Attribute('Script Hash Mismatch Error')
    secondary_credential = Attribute('Secondary Credential')
    service_address = Attribute('Service Address')
    status = Attribute('Status')
    stop_requested = Attribute('Stop Requested')
    template_name_certificate_origin = Attribute('Template Name Certificate Origin')
    use_external_msca_communicator = Attribute('Use External MSCA Communicator')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    CloudApiEndpoint,
    CloudApiOutputModel,
    generate_output,
)
from pyvenafi.cloud.api.models import edgemanagement_service
from uuid import UUID

class _edgemanagement_service:
    def __init__(self, api_obj):
        self.v1 = self._v1(api_obj=api_obj)

    class _v1(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='v1')
            self.billofmaterials = self._billofmaterials(api_obj=self._api_obj, url=f'{self._url}/billofmaterials')
            self.edgeencryptionkeys = self._edgeencryptionkeys(
                api_obj=self._api_obj,
                url=f'{self._url}/edgeencryptionkeys'
            )
            self.edgeinstances = self._edgeinstances(api_obj=self._api_obj, url=f'{self._url}/edgeinstances')
            self.edgeworkers = self._edgeworkers(api_obj=self._api_obj, url=f'{self._url}/edgeworkers')
            self.pairingcodes = self._pairingcodes(api_obj=self._api_obj, url=f'{self._url}/pairingcodes')

        class _billofmaterials(CloudApiEndpoint):
            def get(self):
                class Output(CloudApiOutputModel):
                    BillOfMaterialResponse: edgemanagement_service.BillOfMaterialResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'BillOfMaterialResponse'
                    }
                )

        class _edgeencryptionkeys(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self, edgeInstanceId: UUID):
                data = {
                    'edgeInstanceId': edgeInstanceId,
                }

                class Output(CloudApiOutputModel):
                    EncryptionKeysResponse: edgemanagement_service.EncryptionKeysResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params=data),
                    rc_mapping={
                        200: 'EncryptionKeysResponse'
                    }
                )

            class _ID(CloudApiEndpoint):
                def get(self):
                    class Output(CloudApiOutputModel):
                        EncryptionKeyInformation: edgemanagement_service.EncryptionKeyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'EncryptionKeyInformation'
                        }
                    )

        class _edgeinstances(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self, environmentId: UUID):
                data = {
                    'environmentId': environmentId,
                }

                class Output(CloudApiOutputModel):
                    EdgeInstanceResponse: edgemanagement_service.EdgeInstanceResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params=data),
                    rc_mapping={
                        200: 'EdgeInstanceResponse'
                    }
                )

            class _ID(CloudApiEndpoint):
                def delete(self):
                    class Output(CloudApiOutputModel):
                        EdgeInstanceDeleteResponse: edgemanagement_service.EdgeInstanceDeleteResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._delete(params={}),
                        rc_mapping={
                            204: 'EdgeInstanceDeleteResponse'
                        }
                    )

                def get(self, statusDetails: bool):
                    data = {
                        'statusDetails': statusDetails,
                    }

                    class Output(CloudApiOutputModel):
                        EdgeInstanceInformation: edgemanagement_service.EdgeInstanceInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'EdgeInstanceInformation'
                        }
                    )

                def put(self, EdgeInstanceRequest: edgemanagement_service.EdgeInstanceRequest):
                    data = {**EdgeInstanceRequest.dict()}

                    class Output(CloudApiOutputModel):
                        EdgeInstanceInformation: edgemanagement_service.EdgeInstanceInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            200: 'EdgeInstanceInformation'
                        }
                    )

        class _edgeworkers(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def get(self, edgeInstanceId: UUID):
                data = {
                    'edgeInstanceId': edgeInstanceId,
                }

                class Output(CloudApiOutputModel):
                    EdgeWorkersResponse: edgemanagement_service.EdgeWorkersResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params=data),
                    rc_mapping={
                        200: 'EdgeWorkersResponse'
                    }
                )

            def post(self, EdgeWorkerRequest: edgemanagement_service.EdgeWorkerRequest):
                data = {**EdgeWorkerRequest.dict()}

                class Output(CloudApiOutputModel):
                    EdgeWorkerInformation: edgemanagement_service.EdgeWorkerInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'EdgeWorkerInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.pair = self._pair(api_obj=self._api_obj, url=f'{self._url}/pair')

                def delete(self):
                    class Output(CloudApiOutputModel):
                        EdgeWorkerDeleteResponse: edgemanagement_service.EdgeWorkerDeleteResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._delete(params={}),
                        rc_mapping={
                            204: 'EdgeWorkerDeleteResponse'
                        }
                    )

                def get(self):
                    class Output(CloudApiOutputModel):
                        EdgeWorkerInformation: edgemanagement_service.EdgeWorkerInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'EdgeWorkerInformation'
                        }
                    )

                class _pair(CloudApiEndpoint):
                    def post(self, EdgeWorkerRequest: edgemanagement_service.EdgeWorkerRequest):
                        data = {**EdgeWorkerRequest.dict()}

                        class Output(CloudApiOutputModel):
                            EdgeWorkerInformation: edgemanagement_service.EdgeWorkerInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                200: 'EdgeWorkerInformation'
                            }
                        )

        class _pairingcodes(CloudApiEndpoint):
            def post(self, PairingCodeRequest: edgemanagement_service.PairingCodeRequest):
                data = {**PairingCodeRequest.dict()}

                class Output(CloudApiOutputModel):
                    PairingCodeInformation: edgemanagement_service.PairingCodeInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'PairingCodeInformation'
                    }
                )



from __future__ import annotations

import logging
import warnings
from datetime import datetime

from pyvenafi.logger import api_logger
from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)

class _Authorize(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Authorize')

        vedauth_url = self._url.replace('vedsdk', 'vedauth')
        self.Certificate = self._Certificate(api_obj=api_obj, url=f'{vedauth_url}/Certificate')
        self.Device = self._Device(api_obj=api_obj, url=f'{vedauth_url}/Device')
        self.Integrated = self._Integrated(api_obj=api_obj, url=f'{vedauth_url}/Integrated')
        self.IsAuthServer = self._IsAuthServer(api_obj=api_obj, url=f'{vedauth_url}/IsAuthServer')
        self.Jwt = self._Jwt(api_obj=api_obj, url=f'{vedauth_url}/Jwt')
        self.OAuth = self._OAuth(api_obj=api_obj, url=f'{vedauth_url}/OAuth')
        self.Token = self._Token(api_obj=api_obj, url=f'{vedauth_url}/Token')
        self.Verify = self._Verify(api_obj=api_obj, url=f'{vedauth_url}/Verify')

    def post(self, username, password):
        """
        This POST method is written differently in order to effectively omit the password from being logged.
        """
        warnings.warn(
            'Authorizing to TPP with only a username and password is being deprecated. '
            'Refer to product documentation on using OAuth authentication.'
        )
        body = {
            "Username": username,
            "Password": password
        }

        class Output(WebSdkOutputModel):
            token: str = ApiField(alias='APIKey')
            valid_until: datetime = ApiField(alias='ValidUntil')

        api_logger.debug(f'Authenticating to TPP as "{username}"...')
        with api_logger.suppressed(logging.WARNING):
            response = self._post(data=body)
        api_logger.debug(f'Authenticated as "{username}"!')
        return generate_output(response=response, output_cls=Output)

    class _Certificate(WebSdkEndpoint):
        def post(self, client_id: str, scope: str):
            body = {
                'client_id': client_id,
                'scope'    : scope
            }

            class Output(WebSdkOutputModel):
                access_token: str = ApiField(alias='access_token')
                expires: datetime = ApiField(alias='expires')
                expires_in: int = ApiField(alias='expires_in')
                identity: str = ApiField(alias='identity')
                refresh_token: str = ApiField(alias='refresh_token')
                refresh_until: datetime = ApiField(alias='refresh_until')
                scope: str = ApiField(alias='scope')
                token_type: str = ApiField(alias='token_type')

            api_logger.debug(
                f'Authenticating to TPP OAuth Application "{client_id}" '
                f'with scope "{scope}" using a certificate file...'
            )
            with api_logger.suppressed(logging.WARNING):
                response = self._post(data=body)
            api_logger.debug(f'Authenticated!')
            return generate_output(response=response, output_cls=Output)

    class _Device(WebSdkEndpoint):
        def post(self, client_id: str, scope: str):
            body = {
                'client_id': client_id,
                'scope'    : scope
            }

            class Output(WebSdkOutputModel):
                device_code: str = ApiField(alias='device_code')
                interval: int = ApiField(alias='interval')
                user_code: str = ApiField(alias='user_code')
                verification_uri: str = ApiField(alias='verification_uri')
                verification_url_complete: str = ApiField(alias='verification_uri_complete')
                expires_in: int = ApiField(alias='expires_in')
                expires: datetime = ApiField(alias='expires')

            api_logger.debug(
                f'Authenticating to TPP OAuth Application "{client_id}" '
                f'with scope "{scope}" using a certificate file...'
            )
            with api_logger.suppressed(logging.WARNING):
                response = self._post(data=body)
            api_logger.debug(f'Authenticated!')
            return generate_output(response=response, output_cls=Output)

    class _Integrated(WebSdkEndpoint):
        def post(self, client_id: str, scope: str, state: str = None):
            body = {
                'client_id': client_id,
                'scope'    : scope,
                'state'    : state
            }

            class Output(WebSdkOutputModel):
                access_token: str = ApiField(alias='access_token', default=None)
                expires: datetime = ApiField(alias='expires')
                expires_in: int = ApiField(alias='expires_in')
                identity: str = ApiField(alias='identity')
                refresh_token: str = ApiField(alias='refresh_token')
                refresh_until: datetime = ApiField(alias='refresh_until')
                scope: str = ApiField(alias='scope')
                token_type: str = ApiField(alias='token_type')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _IsAuthServer(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                pass

            return generate_output(output_cls=Output, response=self._get())

    class _Jwt(WebSdkEndpoint):
        def post(self, client_id: str, jwt: str, scope: str):
            body = {
                "client_id": client_id,
                "jwt"      : jwt,
                "scope"    : scope,
            }

            class Output(WebSdkOutputModel):
                access_token: str = ApiField(alias='access_token')
                refresh_token: str = ApiField(alias='refresh_token')
                scope: str = ApiField(alias='scope')
                identity: str = ApiField(alias='identity')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _OAuth(WebSdkEndpoint):
        def post(self, client_id: str, password: str, scope: str, username: str, state: str = None):
            body = {
                'client_id': client_id,
                'password' : password,
                'scope'    : scope,
                'username' : username,
                'state'    : state
            }

            class Output(WebSdkOutputModel):
                access_token: str = ApiField(alias='access_token', default=None)
                expires: datetime = ApiField(alias='expires')
                expires_in: int = ApiField(alias='expires_in')
                identity: str = ApiField(alias='identity')
                refresh_token: str = ApiField(alias='refresh_token')
                refresh_until: datetime = ApiField(alias='refresh_until')
                scope: str = ApiField(alias='scope')
                token_type: str = ApiField(alias='token_type')

            api_logger.debug(
                f'Authenticating to TPP OAuth Application "{client_id}" '
                f'with scope "{scope}" as "{username}"...'
            )
            with api_logger.suppressed(logging.WARNING):
                response = self._post(data=body)
            api_logger.debug(f'Authenticated as {username}!')
            return generate_output(response=response, output_cls=Output)

    class _Token(WebSdkEndpoint):
        def post(self, client_id: str, refresh_token: str, grant_type: str = None, device_code: str = None):
            body = {
                'client_id'    : client_id,
                'refresh_token': refresh_token,
                'grant_type'   : grant_type,
                'device_code'  : device_code
            }

            class Output(WebSdkOutputModel):
                access_token: str = ApiField(alias='access_token')
                expires: datetime = ApiField(alias='expires')
                expires_in: int = ApiField(alias='expires_in')
                refresh_token: str = ApiField(alias='refresh_token')
                refresh_until: datetime = ApiField(alias='refresh_until')
                scope: str = ApiField(alias='scope')
                token_type: str = ApiField(alias='token_type')

            api_logger.debug(f'Authenticating to TPP OAuth application with a refresh token...')
            with api_logger.suppressed(logging.WARNING):
                response = self._post(data=body)
            api_logger.debug(f'Authenticated!')
            return generate_output(response=response, output_cls=Output)

    class _Verify(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                access_issued_on_unix_time: datetime = ApiField(alias='access_issued_on_unix_time')
                application: str = ApiField(alias='application')
                expires_unix_time: datetime = ApiField(alias='expires_unix_time')
                grant_issued_on_unix_time: datetime = ApiField(alias='grant_issued_on_unix_time')
                identity: str = ApiField(alias='identity')
                scope: str = ApiField(alias='scope')
                valid_for: int = ApiField(alias='valid_for')

                @property
                def access_issued_on(self):
                    return self.access_issued_on_unix_time

                @property
                def access_issued_on_ISO8601(self):
                    return self.access_issued_on_unix_time

                @property
                def expires(self):
                    return self.expires_unix_time

                @property
                def expires_ISO8601(self):
                    return self.expires_unix_time

                @property
                def grant_issued_on(self):
                    return self.grant_issued_on_unix_time

                @property
                def grant_issued_on_ISO8601(self):
                    return self.grant_issued_on_unix_time

            return generate_output(response=self._get(), output_cls=Output)



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Permissions(ObjectModel):
    is_associate_allowed: bool = ApiField(alias='IsAssociateAllowed')
    is_auditor: bool = ApiField(alias='IsAuditor')
    is_create_allowed: bool = ApiField(alias='IsCreateAllowed')
    is_delete_allowed: bool = ApiField(alias='IsDeleteAllowed')
    is_manage_permissions_allowed: bool = ApiField(alias='IsManagePermissionsAllowed')
    is_policy_write_allowed: bool = ApiField(alias='IsPolicyWriteAllowed')
    is_private_key_read_allowed: bool = ApiField(alias='IsPrivateKeyReadAllowed')
    is_private_key_write_allowed: bool = ApiField(alias='IsPrivateKeyWriteAllowed')
    is_read_allowed: bool = ApiField(alias='IsReadAllowed')
    is_rename_allowed: bool = ApiField(alias='IsRenameAllowed')
    is_revoke_allowed: bool = ApiField(alias='IsRevokeAllowed')
    is_view_allowed: bool = ApiField(alias='IsViewAllowed')
    is_write_allowed: bool = ApiField(alias='IsWriteAllowed')



from __future__ import annotations

from typing import Union

from pyvenafi.tpp.api.websdk.models.codesign import (
    AppleEnvironment,
    ArchiveFilter,
    CertificateEnvironment,
    CSPEnvironment,
    DotNetEnvironment,
    GPGEnvironment,
    KeyPairEnvironment,
)
from pyvenafi.tpp.features.bases.feature_base import (
    feature,
    FeatureBase,
)

EnvironmentTypes = Union[
    AppleEnvironment,
    CertificateEnvironment,
    CSPEnvironment,
    DotNetEnvironment,
    GPGEnvironment,
    KeyPairEnvironment,
]

@feature('Auditing')
class CodeSignAuditing(FeatureBase):
    def retrieve_archive_entries(
        self,
        archive_filter: ArchiveFilter,
        page_size: int = None,
        page: int = None
    ):
        """
        Retrieve archive entries using an :class:`~.models.codesign.ArchiveFilter`.

        Args:
            archive_filter: Contains filters to apply in retrieval search.
            page_size: Size of paginated result.
            page: Page number of results to retrieve.

        Returns:
             Output dataclass containing :class:`~.models.codesign.ArchiveResults`
        """
        output = self._api.websdk.Codesign.RetrieveArchiveEntries.post(
            archive_filter=archive_filter,
            page_size=page_size,
            page=page
        )
        return output



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class JSSContainerAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "JSS Container"
    contact = Attribute('Contact')
    description = Attribute('Description')
    discovered_by_dn = Attribute('Discovered By DN')
    jss_organization_name = Attribute('JSS Organization Name')
    last_sync = Attribute('Last Sync')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class ScheduleBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Schedule Base"
    blackout = Attribute('Blackout', min_version='21.4')
    days_of_month = Attribute('Days Of Month', min_version='21.4')
    days_of_week = Attribute('Days Of Week', min_version='21.4')
    days_of_year = Attribute('Days Of Year', min_version='21.4')
    hour = Attribute('Hour', min_version='21.4')
    minute = Attribute('Minute', min_version='21.4')
    priority = Attribute('Priority', min_version='21.4')
    reschedule = Attribute('Reschedule', min_version='21.4')
    start_time = Attribute('Start Time', min_version='21.4')
    stop_time = Attribute('Stop Time', min_version='21.4')
    timezone = Attribute('Timezone', min_version='21.4')
    utc = Attribute('UTC', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.encryption_driver import EncryptionDriverAttributes

class NullEncryptionDriverAttributes(EncryptionDriverAttributes, metaclass=IterableMeta):
    __config_class__ = "Null Encryption Driver"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class GroupAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Group"
    assets = Attribute('Assets')
    closed_group = Attribute('Closed Group')
    foreign_security_principal = Attribute('Foreign Security Principal')
    full_name = Attribute('Full Name')
    group_membership = Attribute('Group Membership')
    member = Attribute('Member')
    owner = Attribute('Owner')
    products = Attribute('Products')
    suggested_member = Attribute('Suggested Member')
    team_member_added_by = Attribute('Team Member Added By')
    team_member_added_on = Attribute('Team Member Added On')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes
from pyvenafi.tpp.attributes.client_work_base import ClientWorkBaseAttributes
from pyvenafi.tpp.attributes.x509_certificate_base import X509CertificateBaseAttributes

class ServerCertificateWorkAttributes(
    ApplicationBaseAttributes,
    ClientWorkBaseAttributes,
    X509CertificateBaseAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Server Certificate Work"
    application_attribute = Attribute('Application Attribute')
    application_type = Attribute('Application Type')
    certificate_container = Attribute('Certificate Container')
    friendly_name = Attribute('Friendly Name')
    interval = Attribute('Interval')
    log_threshold = Attribute('Log Threshold')
    naming_pattern = Attribute('Naming Pattern')
    nix_key_store = Attribute('Nix Key Store')
    nix_private_key = Attribute('Nix Private Key')
    path_type = Attribute('Path Type')
    private_key_trustee = Attribute('Private Key Trustee')
    win_key_store = Attribute('Win Key Store')
    win_private_key = Attribute('Win Private Key')



from __future__ import annotations

from datetime import datetime

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)
from pyvenafi.tpp.api.websdk.models.resultcodes import ResultCodes

class Result(ObjectModel):
    code: int = ApiField()

    @property
    def workflow_result(self) -> str:
        return ResultCodes.Client.get(self.code, 'Unknown')

class Details(ObjectModel):
    approval_explanation: str = ApiField(alias='ApprovalExplanation')
    approval_from: str = ApiField(alias='ApprovalFrom')
    approvers: list[str] = ApiField(alias='Approvers', default_factory=list)
    blocking: str = ApiField(alias='Blocking')
    created: datetime = ApiField(alias='Created')
    issued_due_to: str = ApiField(alias='IssuedDueTo')
    status: str = ApiField(alias='Status')
    updated: datetime = ApiField(alias='Updated')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ACMEChallengeAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "ACME Challenge"
    acme_challenge_token = Attribute('ACME Challenge Token', min_version='21.4')
    acme_challenge_validated_on = Attribute('ACME Challenge Validated On', min_version='21.4')
    status = Attribute('Status', min_version='21.4')
    type = Attribute('Type', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.code_signing_environment_template_base import CodeSigningEnvironmentTemplateBaseAttributes

class CodeSigningCertificateEnvironmentTemplateAttributes(
    CodeSigningEnvironmentTemplateBaseAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Code Signing Certificate Environment Template"
    certificate_authority = Attribute('Certificate Authority', min_version='21.4')
    certificate_container_dn = Attribute('Certificate Container DN', min_version='21.4')
    city = Attribute('City', min_version='21.4')
    country = Attribute('Country', min_version='21.4')
    key_algorithm = Attribute('Key Algorithm', min_version='21.4')
    key_storage_location = Attribute('Key Storage Location', min_version='21.4')
    organization = Attribute('Organization', min_version='21.4')
    organizational_unit = Attribute('Organizational Unit', min_version='21.4')
    state = Attribute('State', min_version='21.4')
    x509_subject = Attribute('X509 Subject', min_version='21.4')
    x509_subjectaltname_rfc822 = Attribute('X509 SubjectAltName RFC822', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class LogHeartbeatAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Log Heartbeat"
    event = Attribute('Event')
    rule = Attribute('Rule')
    timeout = Attribute('Timeout')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.x509_certificate import X509CertificateAttributes

class X509CodeSigningCertificateAttributes(X509CertificateAttributes, metaclass=IterableMeta):
    __config_class__ = "X509 Code Signing Certificate"
    owner = Attribute('Owner', min_version='22.1')
    pending_certificate_vault_id = Attribute('Pending Certificate Vault Id', min_version='23.3')
    pending_private_key_vault_id = Attribute('Pending Private Key Vault Id', min_version='23.3')
    pending_public_key_vault_id = Attribute('Pending Public Key Vault Id', min_version='23.3')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class EngineUpgradeAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Engine Upgrade"
    engine_id = Attribute('Engine Id')
    start_time = Attribute('Start Time')
    stop_time = Attribute('Stop Time')
    warning_count = Attribute('Warning Count')



# All of these attributes are defined in _about.py, but it must be imported this way.
__project_name__: str = ''
__version__: str = ''
__author__: str = ''
__author_email__: str = ''
__project_url__: str = ''
exec(open('pyvenafi/_about.py', 'r').read())

from setuptools import setup, find_packages
import os

PROD_REQUIREMENTS = [
    'isodate>=0.6.0',
    'packaging>=21.3',
    'python-dateutil>=2.8.2',
    'requests>=2.24.0',
    'simplejson>=3.17.6',
    'jsonpickle>=2.1.0',
    'pydantic>=1.9.1,<2.0.0'
]

DEV_REQUIREMENTS = [
    'pyodbc==4.0.30',
    'lxml==4.4.1',
    'pandas==1.3.3',
    'paramiko==2.10.4',
    'autopep8>=1.7.0',
    'twine==4.0.2',
    'keyring==23.0.1',
]

DOC_REQUIREMENTS = [
    'Sphinx>=4.3.2',
    'sphinxcontrib-napoleon~=0.7',
    'sphinx-rtd-theme~=1.0.0',
    'sphinx-rtd-dark-mode~=1.2.4',
    'autodoc-pydantic~=1.7.2'
]

if __name__ == '__main__':
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
        long_description = f.read()

    setup(
        name=__project_name__,
        url=__project_url__,
        version=__version__,
        author=__author__,
        author_email=__author_email__,
        packages=find_packages(include=('pyvenafi*',)),
        package_dir={
            '': '.'
        },
        description='Venafi TPP & Cloud API In Python',
        long_description=long_description,
        long_description_content_type='text/markdown',
        keywords=['tpp', 'venafi', 'tpp', 'trust protection platform', 'vaas', 'venafi as a service', 'venafi cloud'],
        install_requires=PROD_REQUIREMENTS,
        extras_require={
            'dev': DEV_REQUIREMENTS,
            'doc': DOC_REQUIREMENTS,
            'all': DOC_REQUIREMENTS + DEV_REQUIREMENTS,
        },
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3.8',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )



from __future__ import annotations

from datetime import datetime

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Counter(ObjectModel):
    a_name: str = ApiField(alias='AName')
    b_name: str = ApiField(alias='BName')
    c_name: str = ApiField(alias='CName')
    description: str = ApiField(alias='Description')
    disabled: bool = ApiField(alias='Disabled')
    error: str = ApiField(alias='Error')
    ignore_count: bool = ApiField(alias='IgnoreCount')
    name: str = ApiField(alias='Name')
    sensitive: bool = ApiField(alias='Sensitive')
    stats_type: int = ApiField(alias='StatsType')
    value_description: str = ApiField(alias='ValueDescription')

class Key(ObjectModel):
    m_item1: str = ApiField(alias='MItem1')
    m_item2: str = ApiField(alias='MItem2')
    m_item3: str = ApiField(alias='MItem3')

class Value(ObjectModel):
    count: int = ApiField(alias='Count')
    sum_value: int = ApiField(alias='SumValue')
    tag_a: str = ApiField(alias='TagA')
    tag_b: str = ApiField(alias='TagB')
    tag_c: str = ApiField(alias='TagC')
    time_frame: datetime = ApiField(alias='TimeFrame')
    type: int = ApiField(alias='Type')

class Result(ObjectModel):
    key: Key = ApiField(alias='Key')
    value: list[Value] = ApiField(alias='Value', default_factory=list)

class CounterContainer(ObjectModel):
    name: str = ApiField(alias='Name')
    stats_code: int = ApiField(alias='StatsCode')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class LogServerAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Log Server"
    log_application_container = Attribute('Log Application Container')
    log_channel = Attribute('Log Channel')
    log_channel_container = Attribute('Log Channel Container')
    log_notification_container = Attribute('Log Notification Container')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class CodeSigningRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Code Signing Root"
    archive_last_evaluated_on = Attribute('Archive Last Evaluated On', min_version='22.3')
    default_ca_container = Attribute('Default CA Container', min_version='21.4')
    default_certificate_container = Attribute('Default Certificate Container', min_version='21.4')
    default_credential_container = Attribute('Default Credential Container', min_version='21.4')
    disable_jwk_public_key_server = Attribute('Disable JWK Public Key Server', min_version='23.1')
    enforce_group_roles = Attribute('Enforce Group Roles', min_version='21.4')
    environment_create_flow_dn = Attribute('Environment Create Flow DN', min_version='23.3')
    environment_delete_flow_dn = Attribute('Environment Delete Flow DN', min_version='23.3')
    environment_edit_flow_dn = Attribute('Environment Edit Flow DN', min_version='23.3')
    environment_template_update_flow_dn = Attribute('Environment Template Update Flow DN', min_version='21.4')
    flow_instance_macro = Attribute('Flow Instance Macro', min_version='21.4')
    key_storage_location = Attribute('Key Storage Location', min_version='21.4')
    key_use_timeout = Attribute('Key Use Timeout', min_version='21.4')
    max_archive_age_days = Attribute('Max Archive Age Days', min_version='22.3')
    max_environment_count = Attribute('Max Environment Count', min_version='23.1')
    prevent_self_dealing = Attribute('Prevent Self Dealing', min_version='21.4')
    prevent_unlimited_signings = Attribute('Prevent Unlimited Signings', min_version='23.1')
    project_delete_flow_dn = Attribute('Project Delete Flow DN', min_version='21.4')
    project_description_tooltip = Attribute('Project Description Tooltip', min_version='21.4')
    request_in_progress_message = Attribute('Request In Progress Message', min_version='21.4')
    storage_driver = Attribute('Storage Driver', min_version='22.3')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import discovery

class _Discovery(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Discovery')
        self.Import = self._Import(api_obj=api_obj, url=f'{self._url}/Import')

    def Guid(self, guid: str):
        return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

    class _Guid(WebSdkEndpoint):
        def delete(self):
            class Output(WebSdkOutputModel):
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._delete(), output_cls=Output)

    class _Import(WebSdkEndpoint):
        def post(self, endpoints: list[discovery.Endpoint], zone_name: str):
            body = {
                'zoneName' : zone_name,
                'endpoints': endpoints
            }

            class Output(WebSdkOutputModel):
                created_certificates: int = ApiField(alias='createdCertificates')
                created_instances: int = ApiField(alias='createdInstances')
                updated_certificates: int = ApiField(alias='updatedCertificates')
                updated_instances: int = ApiField(alias='updatedInstances')
                warnings: list[str] = ApiField(alias='warnings', default_factory=list)
                zone_name: str = ApiField(alias='zoneName')

            return generate_output(response=self._post(data=body), output_cls=Output)



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import preferences as prefs

class _Preferences(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Preferences')

    def get(self, category: str = None, name: str = None, product: prefs.ProductType = None):
        params = {
            'Category': category,
            'Name'    : name,
            'Product' : product
        }

        class Output(WebSdkOutputModel):
            preferences: list[prefs.Preference] = ApiField(alias='Preferences', default_factory=list)

        return generate_output(output_cls=Output, response=self._get(params=params))

    def post(self, preferences: list[prefs.Preference]):
        body = {
            'Preferences': preferences
        }

        return generate_output(output_cls=WebSdkOutputModel, response=self._post(data=body))

    def delete(self, category: str = None, name: str = None, product: prefs.ProductType = None):
        params = {
            'Category': category,
            'Name'    : name,
            'Product' : product
        }

        return generate_output(output_cls=WebSdkOutputModel, response=self._delete(params=params))



from __future__ import annotations

from pyvenafi.tpp.api.websdk.enums import (
    certificate,
    codesign,
    config,
    config_schema,
    metadata,
    oauth,
    permissions,
    rights,
    secret_store,
    ssh,
)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.log_channel import LogChannelAttributes

class SpiLoggerAttributes(LogChannelAttributes, metaclass=IterableMeta):
    __config_class__ = "Spi Logger"
    uri = Attribute('URI')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)
from pyvenafi.tpp.api.websdk.models.resultcodes import ResultCodes

class Result(ObjectModel):
    code: int = ApiField()

    @property
    def secret_store_result(self) -> str:
        return ResultCodes.SecretStore.get(self.code, 'Unknown')

class TypedNameValues(ObjectModel):
    name: str = ApiField(alias='Name')
    type: str = ApiField(alias='Type')
    value: str = ApiField(alias='Value')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.connection_base import ConnectionBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class JumpServerAttributes(ConnectionBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Jump Server"
    consumers = Attribute('Consumers')
    location = Attribute('Location')
    ssh_connection_string = Attribute('SSH Connection String')
    ssh_version = Attribute('SSH Version')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ApertureConfigurationAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Aperture Configuration"
    applies_to = Attribute('Applies To')
    feedback_url = Attribute('Feedback Url')
    help_text = Attribute('Help Text')
    help_url = Attribute('Help Url')
    lost_and_found = Attribute('Lost and Found')
    timeout = Attribute('Timeout')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.credential_driver_base import CredentialDriverBaseAttributes

class IntegratedCredentialDriverAttributes(CredentialDriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Integrated Credential Driver"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class HashiCorpVaultPKIAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "HashiCorp Vault PKI"
    crl_address = Attribute('CRL Address', min_version='21.4')
    create_certificate_authority = Attribute('Create Certificate Authority', min_version='21.4')
    create_pki_role = Attribute('Create PKI Role', min_version='21.4')
    enhanced_key_usage = Attribute('Enhanced Key Usage', min_version='21.4')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    ocsp_address = Attribute('OCSP Address', min_version='21.4')
    policydn = Attribute('PolicyDN', min_version='21.4')
    role_name = Attribute('Role Name', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class FlowGroupAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Flow Group"
    product_code_description = Attribute('Product Code Description')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    ApiField,
    ObjectModel,
)
from datetime import datetime
from typing import (
    Any,
    List,
    Literal,
)
from uuid import UUID

AnyValue = Any
EdgeStatus = Literal[
    "ACTIVE",
    "LOST_CONNECTION",
    "REGISTERED",
    "PAIRED",
    "LOST_CONNECTION_DURING_INSTALL",
    "INSTALLED",
    "INSTALL_FAILED",
    "UNHEALTHY",
    "ERROR",
    "INSTALLING"
]
EdgeType = Literal[
    "ALL",
    "HUB",
    "SATELLITE"
]
KeyAlgorithm = Literal[
    "RSA",
    "ED25519"
]
ProductEntitlement = Literal[
    "ANY",
    "MIRA",
    "DEVOPS",
    "OUTAGE_DETECTION",
    "CODESIGN"
]
WorkerStatus = Literal[
    "DRAFT",
    "PAIRED",
    "ACTIVE",
    "FAILED",
    "INACTIVE"
]

class BillOfMaterialResponse(ObjectModel):
    billsOfMaterials: List[BillOfMaterialsInformation] = ApiField(alias='billsOfMaterials', default_factory=list)

class BillOfMaterialsInformation(ObjectModel):
    charts: List[Chart] = ApiField(alias='charts', default_factory=list)
    name: str = ApiField(alias='name')
    version: str = ApiField(alias='version')

class Chart(ObjectModel):
    name: str = ApiField(alias='name')
    signature: str = ApiField(alias='signature')
    url: str = ApiField(alias='url')
    version: str = ApiField(alias='version')

class ChartDetails(ObjectModel):
    name: str = ApiField(alias='name')
    version: str = ApiField(alias='version')

class EdgeInstanceConnectionDetails(ObjectModel):
    activeMessageCount: int = ApiField(alias='activeMessageCount')
    edgeInstanceId: UUID = ApiField(alias='edgeInstanceId')
    expiredMessageCount: int = ApiField(alias='expiredMessageCount')
    failedMessageCount: int = ApiField(alias='failedMessageCount')
    lastConnectedDate: datetime = ApiField(alias='lastConnectedDate')
    lastDisconnectedDate: datetime = ApiField(alias='lastDisconnectedDate')
    timestamp: datetime = ApiField(alias='timestamp')

class EdgeInstanceDeleteResponse(ObjectModel):
    name: str = ApiField(alias='name')

class EdgeInstanceHealthDetails(ObjectModel):
    edgeInstanceId: UUID = ApiField(alias='edgeInstanceId')
    installedCharts: List[NamespaceChartDetails] = ApiField(alias='installedCharts', default_factory=list)
    nodeStatus: List[NodeStatusDetails] = ApiField(alias='nodeStatus', default_factory=list)
    systemstatus: List[NamespaceStatusDetails] = ApiField(alias='systemstatus', default_factory=list)
    timestamp: datetime = ApiField(alias='timestamp')

class EdgeInstanceInformation(ObjectModel):
    address: str = ApiField(alias='address')
    clientId: str = ApiField(alias='clientId')
    companyId: UUID = ApiField(alias='companyId')
    deploymentDate: datetime = ApiField(alias='deploymentDate')
    edgeStatus: EdgeStatus = ApiField(alias='edgeStatus')
    edgeType: EdgeType = ApiField(alias='edgeType')
    encryptionKeyDeploymentDate: datetime = ApiField(alias='encryptionKeyDeploymentDate')
    encryptionKeyId: str = ApiField(alias='encryptionKeyId')
    environmentId: UUID = ApiField(alias='environmentId')
    id: UUID = ApiField(alias='id')
    integrationServicesCount: int = ApiField(alias='integrationServicesCount')
    kubernetesVersion: str = ApiField(alias='kubernetesVersion')
    lastSeenOnDate: datetime = ApiField(alias='lastSeenOnDate')
    modificationDate: datetime = ApiField(alias='modificationDate')
    name: str = ApiField(alias='name')
    pairingCodeId: UUID = ApiField(alias='pairingCodeId')
    productEntitlements: List[ProductEntitlement] = ApiField(alias='productEntitlements', default_factory=list)
    reconciliationFailed: bool = ApiField(alias='reconciliationFailed')
    statusDetails: EdgeInstanceStatusDetailsInformation = ApiField(alias='statusDetails')
    version: int = ApiField(alias='version')
    workerStatusDetails: List[EdgeWorkerStatusDetailsInformation] = ApiField(
        alias='workerStatusDetails',
        default_factory=list
    )

class EdgeInstanceRequest(ObjectModel):
    name: str = ApiField(alias='name')

class EdgeInstanceResponse(ObjectModel):
    edgeInstances: List[EdgeInstanceInformation] = ApiField(alias='edgeInstances', default_factory=list)

class EdgeInstanceStatusDetailsInformation(ObjectModel):
    connectionDetails: List[EdgeInstanceConnectionDetails] = ApiField(alias='connectionDetails', default_factory=list)
    healthDetails: List[EdgeInstanceHealthDetails] = ApiField(alias='healthDetails', default_factory=list)

class EdgeWorkerDeleteResponse(ObjectModel):
    id: UUID = ApiField(alias='id')
    pairingCode: str = ApiField(alias='pairingCode')

class EdgeWorkerHealthDetails(ObjectModel):
    age: int = ApiField(alias='age')
    cpuUsage: int = ApiField(alias='cpuUsage')
    edgeInstanceId: UUID = ApiField(alias='edgeInstanceId')
    edgeWorkerId: UUID = ApiField(alias='edgeWorkerId')
    host: str = ApiField(alias='host')
    memoryUsage: int = ApiField(alias='memoryUsage')
    port: int = ApiField(alias='port')
    services: List[WorkerServiceStatusDetails] = ApiField(alias='services', default_factory=list)
    status: str = ApiField(alias='status')
    timestamp: datetime = ApiField(alias='timestamp')

class EdgeWorkerInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    edgeInstanceId: UUID = ApiField(alias='edgeInstanceId')
    environmentId: UUID = ApiField(alias='environmentId')
    host: str = ApiField(alias='host')
    id: UUID = ApiField(alias='id')
    lastSeenOnDate: datetime = ApiField(alias='lastSeenOnDate')
    pairingCode: str = ApiField(alias='pairingCode')
    pairingPublicKey: str = ApiField(alias='pairingPublicKey')
    port: int = ApiField(alias='port')
    status: WorkerStatus = ApiField(alias='status')

class EdgeWorkerRequest(ObjectModel):
    edgeInstanceId: UUID = ApiField(alias='edgeInstanceId')
    host: str = ApiField(alias='host')
    port: int = ApiField(alias='port')

class EdgeWorkerStatusDetailsInformation(ObjectModel):
    edgeWorkerId: UUID = ApiField(alias='edgeWorkerId')
    healthDetails: List[EdgeWorkerHealthDetails] = ApiField(alias='healthDetails', default_factory=list)

class EdgeWorkersResponse(ObjectModel):
    edgeWorkers: List[EdgeWorkerInformation] = ApiField(alias='edgeWorkers', default_factory=list)

class EncryptionKeyInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    id: str = ApiField(alias='id')
    key: str = ApiField(alias='key')
    keyAlgorithm: KeyAlgorithm = ApiField(alias='keyAlgorithm')
    lastBackupDate: datetime = ApiField(alias='lastBackupDate')

class EncryptionKeysResponse(ObjectModel):
    encryptionKeys: List[EncryptionKeyInformation] = ApiField(alias='encryptionKeys', default_factory=list)

class ErrorInformation(ObjectModel):
    args: List[AnyValue] = ApiField(alias='args', default_factory=list)
    code: int = ApiField(alias='code')
    message: str = ApiField(alias='message')

class ErrorResponse(ObjectModel):
    errors: List[ErrorInformation] = ApiField(alias='errors', default_factory=list)

class NamespaceChartDetails(ObjectModel):
    charts: List[ChartDetails] = ApiField(alias='charts', default_factory=list)
    name: str = ApiField(alias='name')

class NamespaceStatusDetails(ObjectModel):
    name: str = ApiField(alias='name')
    pods: List[PodStatusDetails] = ApiField(alias='pods', default_factory=list)
    timestamp: datetime = ApiField(alias='timestamp')

class NodeCondition(ObjectModel):
    Status: str = ApiField(alias='Status')
    Type: str = ApiField(alias='Type')
    status: str = ApiField(alias='status')
    type: str = ApiField(alias='type')

class NodeInfo(ObjectModel):
    kubelet_version: str = ApiField(alias='kubelet_version')

class NodeStatusDetails(ObjectModel):
    conditions: List[NodeCondition] = ApiField(alias='conditions', default_factory=list)
    info: NodeInfo = ApiField(alias='info')

class PairingCodeInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    environmentId: UUID = ApiField(alias='environmentId')
    expirationDate: datetime = ApiField(alias='expirationDate')
    id: UUID = ApiField(alias='id')
    pairingCode: str = ApiField(alias='pairingCode')
    productEntitlements: List[ProductEntitlement] = ApiField(alias='productEntitlements', default_factory=list)
    reuseCount: int = ApiField(alias='reuseCount')

class PairingCodeRequest(ObjectModel):
    environmentId: UUID = ApiField(alias='environmentId')
    expirationDate: datetime = ApiField(alias='expirationDate')
    reuseCount: int = ApiField(alias='reuseCount')

class PodStatusDetails(ObjectModel):
    age: int = ApiField(alias='age')
    cpuUsage: int = ApiField(alias='cpuUsage')
    memoryUsage: int = ApiField(alias='memoryUsage')
    name: str = ApiField(alias='name')
    restartCount: int = ApiField(alias='restartCount')
    status: str = ApiField(alias='status')

class WorkerServiceStatusDetails(ObjectModel):
    name: str = ApiField(alias='name')
    status: str = ApiField(alias='status')

BillOfMaterialResponse.update_forward_refs()
BillOfMaterialsInformation.update_forward_refs()
Chart.update_forward_refs()
ChartDetails.update_forward_refs()
EdgeInstanceConnectionDetails.update_forward_refs()
EdgeInstanceDeleteResponse.update_forward_refs()
EdgeInstanceHealthDetails.update_forward_refs()
EdgeInstanceInformation.update_forward_refs()
EdgeInstanceRequest.update_forward_refs()
EdgeInstanceResponse.update_forward_refs()
EdgeInstanceStatusDetailsInformation.update_forward_refs()
EdgeWorkerDeleteResponse.update_forward_refs()
EdgeWorkerHealthDetails.update_forward_refs()
EdgeWorkerInformation.update_forward_refs()
EdgeWorkerRequest.update_forward_refs()
EdgeWorkerStatusDetailsInformation.update_forward_refs()
EdgeWorkersResponse.update_forward_refs()
EncryptionKeyInformation.update_forward_refs()
EncryptionKeysResponse.update_forward_refs()
ErrorInformation.update_forward_refs()
ErrorResponse.update_forward_refs()
NamespaceChartDetails.update_forward_refs()
NamespaceStatusDetails.update_forward_refs()
NodeCondition.update_forward_refs()
NodeInfo.update_forward_refs()
NodeStatusDetails.update_forward_refs()
PairingCodeInformation.update_forward_refs()
PairingCodeRequest.update_forward_refs()
PodStatusDetails.update_forward_refs()
WorkerServiceStatusDetails.update_forward_refs()



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class MetadataBaseAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Metadata Base"
    allowed_values = Attribute('Allowed Values')
    associated_classes = Attribute('Associated Classes')
    category = Attribute('Category')
    default_values = Attribute('Default Values')
    help_text = Attribute('Help Text')
    label_text = Attribute('Label Text')
    localization = Attribute('Localization')
    mandatory = Attribute('Mandatory')
    not_before = Attribute('Not Before')
    policyable = Attribute('Policyable')
    render_hidden = Attribute('Render Hidden')
    render_read_only = Attribute('Render Read Only')
